"""
Generate test code tool

copyrigth https://github.com/shigeshige/py-ut-generator
"""

from typing import List

from pyutgenerator.objects import MockFunc, ParseFunc

STR_PRE_FUNC = 'test_'

TEMP_IMPORT = """
import pytest
from unittest.mock import patch
from unittest.mock import MagicMock

{}
"""

TEMP_FUNC = """
def {}():
    # plan
{}
    # do
{}
{}
{}
"""

TEMP_FUNC_CHECK = """
    # check
{}
"""
TEMP_FUNC_CHECK_TAB = """
        # check
{}
"""


STR_FROM_IMPORT = 'form {} import {}'

STR_VARIS = '    {} = {}'

STR_RUNS = '    {}.{}({})'
STR_RUNS_RETURN = '    ret = {}.{}({})'

STR_RUNS_PRE = '''    target = {}.{}()
'''

STR_WITH = '    with\\'
STR_MOCK = "            patch('{}') as {}"
STR_MOCK_RETURN = '        {}.return_value = {}'
STR_MOCK_FUNC = '        {}.{} = MagicMock(return_value=None)'

STR_RC = '\n'
STR_ASSERT = '    assert {}'
STR_ASSERT_TAB = '        assert {}'
STR_TAB = '    '


def parse_import(pkg, mdn):
    """
    parse import
    """
    owenr = f'from {pkg} import {mdn}'
    if not pkg:
        owenr = f'import {mdn}'

    return TEMP_IMPORT.format(owenr)


def parse_func(fpo: ParseFunc):
    """
    parse one function.
    """
    mdn = fpo.module_name
    run_txt = ''
    checks = ''

    if fpo.has_return:
        checks = parse_assert(['ret'], fpo.mocks)
    inits = '\n'.join([parse_varis(arg, 'None') for arg in fpo.args])

    if fpo.mocks:
        run_txt = STR_TAB
    if fpo.class_name:
        # call for Class
        if fpo.class_func:
            if fpo.has_return:
                runs = run_txt + \
                    STR_RUNS_RETURN.format(
                        mdn, fpo.class_name + '.' + fpo.name, ', '.join(fpo.args))
            else:
                runs = run_txt + STR_RUNS.format(mdn,
                                                 fpo.class_name + '.' + fpo.name,
                                                 ', '.join(fpo.args))
        else:
            runs = run_txt + STR_RUNS_PRE.format(mdn, fpo.class_name)
            if fpo.has_return:
                runs += run_txt + \
                    STR_RUNS_RETURN.format('target', fpo.name, ', '.join(fpo.args))
            else:
                runs += run_txt + \
                    STR_RUNS.format('target', fpo.name, ', '.join(fpo.args))
    else:
        if fpo.has_return:
            runs = run_txt + STR_RUNS_RETURN.format(mdn, fpo.name, ', '.join(fpo.args))
        else:
            runs = run_txt + STR_RUNS.format(mdn, fpo.name, ', '.join(fpo.args))
    mck = parse_mocks(fpo.mocks)
    mck_ret = parse_mocks_return(fpo.mocks)
    if fpo.mocks:
        txt_cheks = TEMP_FUNC_CHECK_TAB.format(checks)
    else:
        txt_cheks = TEMP_FUNC_CHECK.format(checks)

    return TEMP_FUNC.format(STR_PRE_FUNC + fpo.name, inits, mck, mck_ret, runs) + txt_cheks

def parse_varis(name, value):
    """
    parse variers
    """
    return STR_VARIS.format(name, value)


def parse_mocks_return(mocks: List[MockFunc]):
    """
    m.return_value = None
    """
    txt = []
    for i, moc in enumerate(mocks):
        if moc.has_return:
            txt.append(STR_MOCK_RETURN.format('m' + str(i + 1), 'None'))
            if moc.func_name:
                txt.append(STR_MOCK_FUNC.format(
                    'm' + str(i + 1), moc.func_name))

    return STR_RC.join(txt)


def parse_mocks(mocks: List[MockFunc]):
    """
    parse mocks.
    """
    txt = []
    if not mocks:
        return ''
    txt.append(STR_WITH)
    for i, moc in enumerate(mocks):
        txt.append(STR_MOCK.format(moc.mock_path, 'm' + str(i + 1)))
        if len(mocks) - 1 == i:
            txt[-1] += ':'
        else:
            txt[-1] += ',\\'
    return STR_RC.join(txt)


def parse_assert(asserts, tab=False):
    """
    parse assert.
    """
    if tab:
        return STR_RC.join([STR_ASSERT_TAB.format(asst) for asst in asserts])
    else:
        return STR_RC.join([STR_ASSERT.format(asst) for asst in asserts])



"""
Generate test code tool

copyrigth https://github.com/shigeshige/py-ut-generator
"""
import ast
from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class CallFunc:
    """
    Function class
    """
    module: str
    func_name: str
    has_return: bool = False
    module2: Optional[str] = None
    class_name: Optional[str] = None
    ats = None


@dataclass
class MockFunc:
    """
    Mock Class
    """
    mock_path: str
    has_return: bool = False
    func_name: Optional[str] = None

@dataclass
class ParseFunc:
    """
    Parse Func Obj
    """
    name : str
    t_func: ast.FunctionDef = None
    module_name: str = ''
    pakcage : str = ''
    args : List[str] = None
    calls : List[CallFunc] = None
    has_return: bool  = False
    class_name : str = ''
    mocks : List[MockFunc] = field(default_factory=list)
    class_func : bool = False



"""
Generate test code tool

copyrigth https://github.com/shigeshige/py-ut-generator
定数一覧
"""

AST_EXPR = 'Expr'
AST_FUCNTION = 'FunctionDef'
AST_FOR = 'For'
AST_IF = 'If'
AST_IMPORT_FROM = 'ImportFrom'
AST_IMPORT = 'Import'
AST_MODULE = 'Module'
AST_RETURN = 'Return'
AST_CALL = 'Call'
AST_NAME = 'Name'
AST_ATTRIBUTE = 'Attribute'
AST_ASSIGN = 'Assign'
AST_WHILE = 'While'
AST_WITH = 'With'
AST_CLASS = 'ClassDef'

STR_PRE_FUNC = 'test_'



"""
Generate test code tool

copyrigth https://github.com/shigeshige/py-ut-generator
"""

import ast
from typing import List, Optional

from pyutgenerator import const, files
from pyutgenerator.objects import CallFunc, MockFunc, ParseFunc


def create_ast(file_name):
    """
    parse source file.
    """
    src = files.read_file(file_name)
    if not src:
        return None
    return ast.parse(src, file_name)


def get_function(module) -> List[ast.FunctionDef]:
    """
    get function from module.
    """
    funcs = []
    for stm in module.body:
        if _equals(stm, const.AST_FUCNTION):
            funcs.append(stm)
    return funcs


def get_function_class(module):
    """
    get function from class module.
    """
    funcs = []
    for stm in module.body:
        # class define
        if _equals(stm, const.AST_CLASS):
            for stm2 in stm.body:
                if _equals(stm2, const.AST_FUCNTION):
                    funcs.append((stm2, stm.name))
    return funcs


def has_test_function(test_module, func):
    """
    check test function already has.
    """
    if not test_module:
        return False

    for stm in ast.walk(test_module):
        if _equals(stm, const.AST_FUCNTION):
            if get_test_func(func.name) == stm.name:
                return True
    return False



def get_test_func(func_name):
    """
    test function name.
    """
    return const.STR_PRE_FUNC + func_name


def has_return_val(func):
    """
    返却値があるか？
    """
    for stm in ast.walk(func):
        if _equals(stm, const.AST_RETURN):
            if stm.value:
                return True
    return False


def get_func_arg(func):
    """
    関数の引数取得
    """
    prms = [a.arg for a in func.args.args]
    return prms


def get_calls(func) -> List[CallFunc]:
    """
    関数呼び出しを取得する
    """
    calls = []
    for stm in ast.walk(func):
        if _equals(stm, const.AST_CALL):
            # print('---cal--')
            # call(hoge)
            if _equals(stm.func, const.AST_NAME):
                has_return = _has_return_call(stm, func)
                calls.append(CallFunc('', stm.func.id, has_return))
            # hoge.call(hoge)
            if _equals(
                    stm.func,
                    const.AST_ATTRIBUTE) and _equals(
                        stm.func.value,
                        const.AST_NAME):
                has_return = _has_return_call(stm, func)
                calls.append(CallFunc(stm.func.value.id, stm.func.attr, has_return))
            # hoge.hoge.call(hoge)
            if _equals(
                    stm.func, const.AST_ATTRIBUTE) and _equals(
                        stm.func.value, const.AST_ATTRIBUTE) and _equals(
                        stm.func.value.value, const.AST_NAME):
                has_return = _has_return_call(stm, func)
                calls.append(
                    CallFunc(
                        stm.func.value.attr, stm.func.attr, has_return,
                        stm.func.value.value.id))
    return calls


def get_funcs(module) -> List[ast.FunctionDef]:
    """
    get function list
    """
    return [stm for stm in ast.walk(module) if _equals(stm, const.AST_FUCNTION)]


def _has_return_call(call_obj, func):
    """
    call ast uses return value?
    """
    for stm in ast.walk(func):
        # bbb = aaa()
        if _equals(stm, const.AST_ASSIGN):
            for stm2 in ast.walk(stm.value):
                if stm2 is call_obj:
                    return True
        # bbb(aaa())
        if _equals(stm, const.AST_CALL):
            for stm2 in [j for i in stm.args for j in ast.walk(i)]:
                if stm2 is call_obj:
                    return True
        # if aaa():
        if _equals(stm, const.AST_IF) or _equals(stm, const.AST_WHILE):
            for stm2 in ast.walk(stm.test):
                if stm2 is call_obj:
                    return True
        # with aaa() as bbb
        if _equals(stm, const.AST_WITH):
            for stm2 in ast.walk(stm):
                if stm2 is call_obj:
                    return True
    return False


def _create_mock_func(stm, clf: CallFunc, pkg, mdn) -> Optional[MockFunc]:
    """
    create mocks
    """
    # from xxx import yyy
    if clf.module in list(map(lambda x: x.name, stm.names)):
        return MockFunc(
            stm.module + '.' + clf.module + '.' + clf.func_name, clf.has_return)
    # from xxx.xxx import yyy
    if clf.func_name in list(map(lambda x: x.name, stm.names)):
        if clf.module:
            return MockFunc(
                stm.module + '.' + clf.module + '.' + clf.func_name, clf.has_return)
        else:
            return MockFunc(
                pkg + '.' + mdn + '.' + clf.func_name, clf.has_return)
    return None


def get_mocks(calls: List[CallFunc], module, pkg, mdn):
    """
    呼び出し先のモックを作成する。
    """
    mocks: List[MockFunc] = []
    for clf in calls:
        import_flg = False
        for stm in ast.walk(module):
            if _equals(stm, const.AST_IMPORT):
                if get_import_names(stm) == clf.func_name and not clf.module:
                    # import xxx
                    import_flg = True
                elif get_import_names(stm) == clf.module:
                    if clf.module2:
                        mocks.append(
                            MockFunc(
                                pkg + '.' + mdn + '.' + clf.module2,
                                clf.has_return, clf.func_name))
                    else:
                        mocks.append(MockFunc(clf.module + '.' + clf.func_name, clf.has_return))
                    import_flg = True
            if _equals(stm, const.AST_IMPORT_FROM):
                mck = _create_mock_func(stm, clf, pkg, mdn)
                if mck is not None:
                    mocks.append(mck)
        if import_flg:
            continue
        for stm in get_function(module):
            # def xxx():
            if not clf.module and clf.func_name == stm.name:
                mocks.append(MockFunc(pkg + '.' + mdn + '.' + clf.func_name, clf.has_return))

    return mocks


def get_import_names(stm : ast.Import):
    """
    join import names.
    """
    return '.'.join([x.name for x in stm.names])


def _equals(stm, class_name) -> bool:
    """
    check class name.
    """
    return stm.__class__.__name__ == class_name


def make_func_obj(t_func : ast.FunctionDef, package, module_name, module, class_name='') -> ParseFunc:
    """
    関数解析
    """

    pfo = ParseFunc(
        t_func.name, t_func,
        module_name,
        package,
        get_func_arg(t_func),
        get_calls(t_func),
        has_return_val(t_func))
    pfo.class_func = len(
        list(
            filter(
                lambda x: getattr(x, 'id', None) in [
                    'staticmethod', 'classmethod'],
                t_func.decorator_list))) != 0
    pfo.mocks = get_mocks(pfo.calls, module, package, module_name)
    pfo.class_name = class_name
    return pfo



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyutgenerator']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['pyutgen = pyutgenerator.run:main']}

setup_kwargs = {
    'name': 'pyutgenerator',
    'version': '0.4.1',
    'description': 'python ut test code generator.',
    'long_description': None,
    'author': 'shigeshige',
    'author_email': '5540474+shigeshige@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/shigeshige/py-ut-generator',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Generate test code tool

copyrigth https://github.com/shigeshige/py-ut-generator
"""

import os
import pathlib


def read_file(file_name):
    """
    read utf8 txt file.
    """
    if not os.path.exists(file_name):
        return None

    with open(file_name, 'r', encoding='utf-8') as f:
        return f.read()


def write_file(file_name, txt, append=False):
    """
    write txt file.
    and mkdir directory
    """
    mode = 'w'
    if append and os.path.exists(file_name):
        mode = 'a'
    if not txt:
        print('No output')
        return
    dirs = os.path.dirname(file_name)
    os.makedirs(dirs, exist_ok=True)
    with open(file_name, mode=mode, encoding='utf8') as fip:
        fip.write(txt)


def get_package_moduel(file_name):
    """
    get package name and module name from file name.
    """
    modu = os.path.splitext(os.path.basename(file_name))[0]
    pth = pathlib.Path(file_name).absolute()
    pkg = os.path.dirname(str(pth.relative_to(pathlib.Path.cwd())))
    pkg = pkg.replace('\\', '.')
    pkg = pkg.replace('/', '.')
    if os.path.basename(file_name) == pkg:
        pkg = ''
    return pkg, modu


def get_test_file_name(pkg, mdn):
    """
    get test file name from input file.
    """
    return 'tests/' + pkg.replace('.', '/') + '/test_' + mdn + '.py'



"""
Generate test code tool


copyrigth https://github.com/shigeshige/py-ut-generator
"""
import argparse

from pyutgenerator import ast_util, files, templates


def parse_file(file_name, renew=False):
    """
    generate test code.
    """
    module = ast_util.create_ast(file_name)
    if not module:
        print('File not Found :' + str(file_name))
        return
    t_funcs = ast_util.get_function(module)
    t_funcs_cls = ast_util.get_function_class(module)
    pkg, mdn = files.get_package_moduel(file_name)
    # test file name
    t_file = files.get_test_file_name(pkg, mdn)
    old_test = ast_util.create_ast(t_file)
    append = old_test is not None
    if renew:
        append = False
        old_test = None
    if append:
        ttt = ''
    else:
        ttt = templates.parse_import(pkg, mdn)

    for t_func in t_funcs:
        if ast_util.has_test_function(old_test, t_func):
            continue
        fpo = ast_util.make_func_obj(t_func, pkg, mdn, module)
        ttt += templates.parse_func(fpo)

    for t_func, clazz in t_funcs_cls:
        if ast_util.has_test_function(old_test, t_func):
            continue
        fpo = ast_util.make_func_obj(t_func, pkg, mdn, module, clazz)
        ttt += templates.parse_func(fpo)

    files.write_file(t_file, ttt, append)


def main():
    """
    main function.
    """
    pas = argparse.ArgumentParser()
    pas.add_argument("filename", help="input python filename")
    pas.add_argument(
        "--ovewrite", help="overwrite test code", action="store_true")
    args = pas.parse_args()
    print(args.filename)
    print(args.ovewrite)
    parse_file(args.filename, args.ovewrite)
    return


if __name__ == "__main__":
    main()
